/*****************************************************************************
 *
 * FILE:	api2.c
 * DESCRIPTION:	extended API for WKF
 * DATE:	Wed, Jun 13 2001
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 2001 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: api2.c,v 1.1.2.7 2002/10/02 15:50:14 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "kanji.h"
#include "functions.h"


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	ISO_8859_1_Q	"=?ISO-8859-1?Q?"
#define	ISO_2022_JP_B	"=?ISO-2022-JP?B?"
#define	ISO_2022_JP_Q	"=?ISO-2022-JP?Q?"
#define	JAPANESE_EUC_B	"=?JAPANESE_EUC?B?"
#define	SHIFT_JIS_B	"=?SHIFT_JIS?B?"

#define	length(x)	(sizeof(x) - 1)


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static int encodeBase64(const String, size_t, String, size_t);
static int decodeBase64(const String, size_t, String, size_t);
static bool_t	skipChar(char);


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
/*
 * bit sequence for mask
 *
 * 0xfc = 11111100
 * 0x03 = 00000011
 * 0xf0 = 11110000
 * 0x0f = 00001111
 * 0xc0 = 11000000
 * 0x3f = 00111111
 * 0x30 = 00110000
 * 0x3c = 00111100
 */

/*
 * Base64 alphabet
 */
const char	base64[] =
	"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * Function Name:	wkfEncodeBase64String
 * Description:		encode a text string to a base64 string
 * Arguments:		string, received buffer, received buffer size
 * Return Value:	Successful: encoded string byte, otherwise: -1
 */
int
wkfEncodeBase64String(str, rbuf, rsize)
	const String	str;
	String		rbuf;	/* received base64 encoded string */
	size_t		rsize;	/* received buffer size */
{
  size_t	len = strlen(str);
  return encodeBase64(str, len, rbuf, rsize);
}

/*
 * Function Name:	wkfEncodeBase64Bin
 * Description:		encode a data string to a base64 string
 * Arguments:		data, data length, received buffer, received buffer size
 * Return Value:	Successful: encoded string byte, otherwise: -1
 */
int
wkfEncodeBase64Bin(data, len, rbuf, rsize)
	const String	data;
	size_t		len;
	String		rbuf;	/* received base64 encoded string */
	size_t		rsize;	/* received buffer size */
{
  return encodeBase64(data, len, rbuf, rsize);
}

/*
 * Function Name:	encodeBase64
 * Description:		encode a data to a base64 string
 * Arguments:		data, data length, received buffer, received buffer size
 * Return Value:	Successful: encoded string byte, otherwise: -1
 */
static int
encodeBase64(data, len, rbuf, rsize)
	const String	data;
	size_t		len;
	String		rbuf;	/* received base64 encoded string */
	size_t		rsize;	/* received buffer size */
{
  int	i, j;

  if (rsize < (int)((double)len * 1.33) + 1) {
    return -1;
  }

  /*
   * encode 3-bytes (24-bits) at a time
   */
  memset(rbuf, 0, rsize);
  for (i = j = 0; i < len - (len % 3); i += 3, j += 4) {
    rbuf[j]   = base64[( data[i]   & 0xfc) >> 2];
    rbuf[j+1] = base64[((data[i]   & 0x03) << 4) | ((data[i+1] & 0xf0) >> 4)];
    rbuf[j+2] = base64[((data[i+1] & 0x0f) << 2) | ((data[i+2] & 0xc0) >> 6)];
    rbuf[j+3] = base64[( data[i+2] & 0x3f)];
  }

  i = len - (len % 3);	/* rest size */
  switch (len % 3) {
    case 2:	/* one character padding */
      rbuf[j]	= base64[( data[i]   & 0xfc) >> 2];
      rbuf[j+1] = base64[((data[i]   & 0x03) << 4) | ((data[i+1] & 0xf0) >> 4)];
      rbuf[j+2] = base64[( data[i+1] & 0x0f) << 2];
      rbuf[j+3] = base64[64];	/* Pad	*/
      break;
    case 1:	/* two character padding */
      rbuf[j]	= base64[(data[i] & 0xfc) >> 2];
      rbuf[j+1]	= base64[(data[i] & 0x03) << 4];
      rbuf[j+2] = base64[64];	/* Pad	*/
      rbuf[j+3] = base64[64];	/* Pad	*/
      break;
  }
  rbuf[j+4] = EOL;

  return j;
}

/*
 * Function Name:	wkfDecodeBase64String
 * Description:		decode a base64 string to a normal text string
 * Arguments:		string, received buffer, received buffer size
 * Return Value:	Successful: decoded string byte, otherwise: -1
 */
int
wkfDecodeBase64String(str, rbuf, rsize)
	const String	str;
	String		rbuf;	/* received base64 decoded string */
	size_t		rsize;	/* received buffer size */
{
  size_t	len = strlen(str);
  return decodeBase64(str, len, rbuf, rsize);
}

#if	0
/*
 * Function Name:	wkfDecodeBase64Bin
 * Description:		decode a base64 string to a data string
 * Arguments:		data, data length, received buffer, received buffer size
 * Return Value:	Successful: decoded string byte, otherwise: -1
 */
int
wkfDecodeBase64Bin(data, len, rbuf, rsize)
	const String	data;
	size_t		len;
	String		rbuf;	/* received base64 decoded string */
	size_t		rsize;	/* received buffer size */
{
  return decodeBase64(data, len, rbuf, rsize);
}
#endif

/*
 * Function Name:	decodeBase64
 * Description:		decode a base64 string to a data string
 * Arguments:		data, data length received buffer, received buffer size
 * Return Value:	Successful: decoded string byte, otherwise: -1
 */
static int
decodeBase64(data, len, rbuf, rsize)
	const String	data;
	size_t		len;
	String		rbuf;	/* received base64 decoded string */
	size_t		rsize;	/* received buffer size */
{
#define	VAL(x)	(data[(x)] == '=' ? 0 : strchr(base64, data[(x)]) - base64)
  int	i, j;

  if (rsize < (int)((double)len * 0.75) + 1) {
    return -1;
  }

  /*
   * work on 4-words (24-bits) at a time
   */
  memset(rbuf, 0, rsize);
  for (i = j = 0; i < len; i += 4, j += 3) {
    rbuf[j]   = (VAL(i) << 2) | ((VAL(i+1) & 0x30) >> 4);
    rbuf[j+1] = ((VAL(i+1) & 0x0f) << 4) | ((VAL(i+2) & 0x3c) >> 2); 
    rbuf[j+2] = ((VAL(i+2) & 0x03) << 6) | (VAL(i+3) & 0x3f);
  }
  for (i = j; rbuf[i] == NULL && i > j - 3; i--) {
    ;	/* remove padding data */
  }
  j = i + 1;

  return j;
}

/*****************************************************************************/

/*
 * Function Name:	wkfDecodeBase64File
 * Description:		decode a base64 encoded file
 * Arguments:		input file pointer, output file pointer
 * Return Value:	none
 */
void
wkfDecodeBase64File(fin, fout)
	register FILE *	fin;
	register FILE *	fout;
{
#define	VAL2(x)	((x) == '=' ? 0 : strchr(base64, (x)) - base64)
  int	c1, c2, c3, c4;

  /*
   * work on 4-words (24-bits) at a time
   */
  while (!feof(fin)) {
    do { c1 = fgetc(fin); } while (skipChar(c1));
    do { c2 = fgetc(fin); } while (skipChar(c2));
    do { c3 = fgetc(fin); } while (skipChar(c3));
    do { c4 = fgetc(fin); } while (skipChar(c4));
    if (c1 == EOF || c2 == EOF || c3 == EOF || c4 == EOF) {
      return;
    }
    fputc((VAL2(c1) << 2) | ((VAL2(c2) & 0x30) >> 4), fout);
    fputc(((VAL2(c2) & 0x0f) << 4) | ((VAL2(c3) & 0x3c) >> 2), fout); 
    fputc(((VAL2(c3) & 0x03) << 6) | (VAL2(c4) & 0x3f), fout);
  }
}

static bool_t
skipChar(c)
	char	c;
{
  switch (c) {
    case SPC:
    case TAB:
    case CR:
    case LF:
    case EOL:
      return TRUE;
    default:
      return FALSE;
  }
}

/*****************************************************************************/

/*
 * Function Name:	wkfEncodeQuotedPrintableString
 * Description:		encode a string to a quoted-printable string
 * Arguments:		string, received buffer, received buffer size
 * Return Value:	Successful: encoded string byte, otherwise: -1
 */
int
wkfEncodeQuotedPrintableString(str, rbuf, rsize)
	const String	str;
	String		rbuf;	/* received quoted-printable encoded string */
	size_t		rsize;	/* received buffer size */
{
  static char	HEX[]	= "0123456789ABCDEF";
  size_t	len	= 0;
  String	os	= str;
  String	s	= rbuf;
  char		c;
  int		i	= 0;

  len = strlen(str);
  if (rsize < (int)((double)len * 1.33) + 1) {
    return -1;
  }

  memset(rbuf, 0, rsize);
  while ((c = *os) != EOL) {
    /* White Space */
    if (c >= 9 && c <= 32) {
      *s++ = '='; i++;
      *s++ = '2'; i++;
      *s++ = '0'; i++;
      os++;
    }
    /* Literal representation */
    else if ((c >= 33 && c <= 60) || (c >= 62 && c <= 126)) {
      *s++ = c; i++;
      os++;
    }
    /* equal character */
    else if (c == 61) {
      *s++ = '='; i++;
      *s++ = '3'; i++;
      *s++ = 'D'; i++;
      os++;
    }
    /* General 8bit representation */
    else {
      *s++ = '='; i++;
      *s++ = HEX[(c >> 4) & 0x0f]; i++;
      *s++ = HEX[c & 0x0f]; i++;
      os++;
    }
  }

  return i;
}

/*
 * Function Name:	wkfDecodeQuotedPrintableString
 * Description:		decode a quoted-printable string to a normal string
 * Arguments:		string, received buffer, received buffer size
 * Return Value:	Successful: decoded string byte, otherwise: -1
 */
int
wkfDecodeQuotedPrintableString(str, rbuf, rsize)
	const String	str;
	String		rbuf;	/* received quoted-printable decoded string */
	size_t		rsize;	/* received buffer size */
{
  size_t		len	= 0;
  register String	os	= str;
  register String	s	= rbuf;
  char			c, c1;
  int			i	= 0;

  len = strlen(str);
  if (rsize < (int)((double)len * 0.75) + 1) {
    return -1;
  }

  memset(rbuf, 0, rsize);
  while ((c = *os) != EOL) {
    if (c == '=') {
      os++;
      if ((c = *os++) && (c1 = *os++)) {
	if (c == CR || c == LF) {
	  continue;
	}
	*s++ =  ((c  >= 'A' ? ((c  & 0xdf) - 'A') + 10 : c  - '0') * 16) +
		((c1 >= 'A' ? ((c1 & 0xdf) - 'A') + 10 : c1 - '0')); i++;
      }
      else {
	return i;
      }
    }
    else {
      *s++ = c; i++;
      os++;
    }
  }

  return i;
}

/*
 * Function Name:	wkfQuotedPrintableBase64File
 * Description:		decode a quoted-printable encoded file
 * Arguments:		input file pointer, output file pointer
 * Return Value:	none
 */
void
wkfDecodeQuotedPrintableFile(fin, fout)
	register FILE *	fin;
	register FILE *	fout;
{
  int	c, c1;

  while (!feof(fin)) {
    c = fgetc(fin);
    if (c == '=') {
      c	 = fgetc(fin);
      c1 = fgetc(fin);
      if (c == EOF || c1 == EOF) {
	fflush(fout);
	return;
      }
      if (c == CR || c == LF || c == EOL) {
	ungetc(c1, fin);
	continue;
      }
      fputc(((c  >= 'A' ? ((c  & 0xdf) - 'A') + 10 : c  - '0') * 16) +
	    ((c1 >= 'A' ? ((c1 & 0xdf) - 'A') + 10 : c1 - '0')), fout);
    }
    else {
      fputc(c, fout);
    }
  }
  fflush(fout);
}

/*****************************************************************************/

/*
 * Function Name:	wkfDecodeMimeString
 * Description:		decode a MIME encoded string
 * Arguments:		string, received buffer, received buffer size, kcode
 * Return Value:	Successful: CONV_OK, otherwise CONV_ERR
 */
conv_t
wkfDecodeMimeString(str, rbuf, rsize, kout)
	const String	str;
	String		rbuf;	/* received base64 decoded string */
	size_t		rsize;	/* received buffer size */
	kcode_t		kout;
{
  String		mstr	= NULL;	/* mime encoded string (copy of str) */
  size_t		len	= 0;
  int			n	= -1;
  register String	p	= NULL;	/* pointer for str */
  register String	q	= NULL;	/* pointer for recv */
  bool_t		f_spc	= FALSE;/* white space flag */
  bool_t		f_cnt	= FALSE;/* just before mime decoded flag */

  if (str == NULL) {
    return CONV_ERR;
  }

  mstr = strdup(str);
  if (mstr == NULL) {
    return CONV_ERR;
  }

  /* deocde mime string */
  memset(rbuf, 0, rsize);
  for (len = 0, p = mstr, q = rbuf, f_spc = FALSE; *p != EOL && len < rsize; ) {
    if (*p == '=') {
      String	end	 = NULL;
      bool_t	f_base64 = FALSE;
      bool_t	f_quoted = FALSE;

      if (strncasecmp(p, ISO_2022_JP_B, length(ISO_2022_JP_B)) == 0) {
	f_base64 = TRUE;
	p	+= length(ISO_2022_JP_B);
      }
      else if (strncasecmp(p, JAPANESE_EUC_B, length(JAPANESE_EUC_B)) == 0) {
	f_base64 = TRUE;
	p	+= length(JAPANESE_EUC_B);
      }
      else if (strncasecmp(p, SHIFT_JIS_B, length(SHIFT_JIS_B)) == 0) {
	f_base64 = TRUE;
	p	+= length(SHIFT_JIS_B);
      }
      else if (strncasecmp(p, ISO_8859_1_Q, length(ISO_8859_1_Q)) == 0) {
	f_quoted = TRUE;
	p	+= length(ISO_8859_1_Q);
      }
      else if (strncasecmp(p, ISO_2022_JP_Q, length(ISO_2022_JP_Q)) == 0) {
	f_quoted = TRUE;
	p	+= length(ISO_2022_JP_Q);
      }
      else {
      }

      if (f_base64 || f_quoted) {
	end = strstr(p, "?=");
	if (end == NULL) {	/* invalid MIME encoding */
	  free((String)mstr);
	  return CONV_OK;
#if	0
	  return CONV_ERR;	/* if you don't allow any errors. */
#endif
	}
	*end = EOL;

	if (f_spc) {	/* reset */
	  if (f_cnt == FALSE) {
	    *q++ = SPC;
	    len++;
	  }
	  f_spc = FALSE;
	}
      }

      if (f_base64) {
	n = wkfDecodeBase64String(p, q, rsize - len);
	if (n == -1) {
	  free((String)mstr);
	  return CONV_ERR;
	}
	p = end + 2;
	q += n;
	len += n;

	f_base64 = FALSE;	/* clear */
	f_cnt	 = TRUE;

	continue;
      }

      if (f_quoted) {
	n = wkfDecodeQuotedPrintableString(p, q, rsize - len);
	if (n == -1) {
	  free((String)mstr);
	  return CONV_ERR;
	}
	p = end + 2;
	q += n;
	len += n;

	f_quoted = FALSE;	/* clear */
	f_cnt	 = TRUE;

	continue;
      }
    }
    else if (*p == SPC) {
      f_spc = TRUE;
      p++;
      continue;
    }
    if (f_spc) {
      f_spc = FALSE;
      *q++  = SPC;
      len++;
    }
    *q++ = *p++;
    len++;

    f_cnt = FALSE;
  }
  free((String)mstr);

  /* convert kanji code */
  if (kout == KC_JIS || kout == KC_EUC || kout == KC_SJIS) {
    mstr = strdup(rbuf);
    if (mstr == NULL) {
      return CONV_ERR;
    }

    memset(rbuf, 0, rsize);
    switch (wkfConvertKanjiCodeOfString(KC_UNKNOWN, mstr, kout, rbuf, rsize)) {
      case CONV_ERR:
	free((String)mstr);
	return CONV_ERR;
      case CONV_NO:
      case CONV_OK:
      default:
    }
    free((String)mstr);
  }

  return CONV_OK;
}

/*
 * Function Name:	wkfStringToHexString
 * Description:		convert the string to the hex string
 * Arguments:		string, received buffer, received buffer size
 * Return Value:	CONV_OK, CONV_NO or CONV_ERR
 */
conv_t
wkfStringToHexString(str, rbuf, rsize)
	const String	str;
	String		rbuf;	/* received base64 decoded string */
	size_t		rsize;	/* received buffer size */
{
  static char	hex[]	= "0123456789abcdef";
  String	sjis	= NULL;	/* sjis string */
  size_t	len	= 0;
  String	p	= NULL;
  String	q	= NULL;

  /* convert str to SJIS kanji code */
  len	= strlen(str) + 1;
  sjis	= (String)calloc(len, sizeof(char));
  if (sjis == NULL) {
    return CONV_ERR;
  }
  switch (wkfConvertKanjiCodeOfString(KC_UNKNOWN, str, KC_SJIS, sjis, len)) {
    case CONV_ERR:
      free((String)sjis);
      return CONV_ERR;
    case CONV_NO:
      return CONV_NO;
    case CONV_OK:
    default:
      break;
  }

  /* convert sjis to hex string */
  for (p = sjis, q = rbuf; *p != EOL && q - rbuf < rsize - 1; ) {
    if (IsASCII(*p)) {
      *q++ = *p++;
    }
    else if (IsSJIS1(*p) && IsSJIS2(*(p+1))) {
      *q++ = ':';
      *q++ = hex[(*p >> 4) & 0x0f];
      *q++ = hex[*p++ & 0x0f];
      *q++ = ':';
      *q++ = hex[(*p >> 4) & 0x0f];
      *q++ = hex[*p++ & 0x0f];
    }
    else {
      *q++ = *p++;
    }
  }
  *++q = EOL;

  free((String)sjis);

  return CONV_OK;
}

/*
 * Function Name:	wkfMimeStringToHexString
 * Description:		decode a MIME encoded string and convert hex string
 * Arguments:		string, received buffer, received buffer size
 * Return Value:	CONV_OK, CONV_NO or CONV_ERR
 */
conv_t
wkfMimeStringToHexString(str, rbuf, rsize)
	const String	str;
	String		rbuf;	/* received base64 decoded string */
	size_t		rsize;	/* received buffer size */
{
  String	mstr = NULL;
  conv_t	ret;

  memset(rbuf, 0, rsize);
  if (wkfDecodeMimeString(str, rbuf, rsize, KC_UNKNOWN) == CONV_ERR) {
    return CONV_ERR;
  }

  mstr = strdup(rbuf);
  if (mstr == NULL) {
    return CONV_ERR;
  }

  ret = wkfStringToHexString(mstr, rbuf, rsize);
  free((String)mstr);

  return ret;
}

