/*****************************************************************************
 *
 * FILE:	wkf.c
 * DESCRIPTION:	WALL's Kanji Filter
 * DATE:	Sat, Jan 15 2000
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.com
 * URL:		http://www.MysticWALL.com/cover.html
 * COPYRIGHT:	(C) 1997-2000 졿Kouichi ABE (WALL), All rights reserved.
 * $Id: wkf.c,v 1.1.2.3.2.3 2002/10/02 15:50:14 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#ifndef _WIN32
#include <unistd.h>
#endif	/* _WIN32 */
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(_WIN32) || defined(__EMX__)
#include <sys/utime.h>
#include <fcntl.h>
#else
#include <utime.h>
#endif	/* _WIN32 __EMX__ */
#include <signal.h>

#define	MAIN_MODULE
#include "wkf.h"
#include "version.h"
#undef	MAIN_MODULE


/*****************************************************************************
 *
 *	definition of macros and structures
 *
 *****************************************************************************/
#define	FILE_LEN	256	/* max length of filename	*/
#define	EOL		'\0'	/* End of Line */

typedef struct _option {	/* command option type */
  bool_t	checkcode;	/* flag to check kanji code */
  bool_t	showfname;	/* flag to show filename */
  bool_t	mimedecode;	/* flag to decode mime encoding string */
  bool_t	backup;		/* flag to do backup */
  bool_t	savetstamp;	/* flag to save the time-stamp of file */
  bool_t	lineconv;	/* flag to convert string with each line */
  bool_t	lineend;	/* type of the code of line end	*/
} Option;

typedef enum {	/* error code type */
  ERR_NO_CONVERT_OPTION,	/* no option to convert kanji code */
  ERR_MULTIPLE_CONVERT_OPTION,	/* multiple option to convert */
  ERR_MULTIPLE_INPUT_CODE_OPTION, /* multiple option of input kanji code */
  ERR_MULTIPLE_LINE_END_OPTION	/* multiple option of the code of line end */
} error_t;


/*****************************************************************************
 *
 *	declaration of local variables
 *
 *****************************************************************************/
const char	Program[] = "wkf";


/*****************************************************************************
 *
 *	declaration of local functions
 *
 *****************************************************************************/
static conv_t	convertKanjiCodeInFile(kcode_t, FILE *, kcode_t, FILE *,
				       Option *);
static conv_t	convertKanjiCodeOfEachLineInFile(kcode_t, FILE *,
						 kcode_t, FILE *, bool_t);
static void	savePermission(String, String);
static void	saveTimeStamp(String, String);
static String	tmpName(String);
static String	backupName(String);
static void	printErrorMessage(error_t);
static void	copyright(void);
static void	usage(void);


/*****************************************************************************
 *
 *	definition of functions
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	convertKanjiCodeInFile
 * DESCRIPTION:		convert kanji code in a file
 * ARGUMENTS:		input kanji code, input file pointer,
 *			output kanji code, output file pointer, command option
 * RETURN VALUE:	Successful: CONV_OK, otherwise: CONV_ERR
 */
static conv_t
convertKanjiCodeInFile(kin, fin, kout, fout, opt)
	kcode_t		kin;
	register FILE *	fin;
	kcode_t		kout;
	register FILE *	fout;
	Option *	opt;
{
  conv_t	ret;

#if defined(_WIN32) || defined(__EMX__)
  /*  D,M,U Ū˻ꤷʤ CRLF */
  _setmode(fileno(fin), O_BINARY);  /* ϤϾ binary open */
  if (opt->lineend == FALSE) {
    wkfSetLineEndCode(LE_CRLF);
    _setmode(fileno(fout), O_BINARY);
  }
#endif	/* _WIN32 __EMX__ */
  setvbuf(fin,  (char *)NULL, _IOFBF, 0);
  setvbuf(fout, (char *)NULL, _IOFBF, 0);
  if (opt->lineconv) {
    ret = convertKanjiCodeOfEachLineInFile(kin, fin, kout, fout,
					   opt->mimedecode);
  }
  else {
    ret = wkfConvertKanjiCodeOfFile(kin, fin, kout, fout);
  }
  fflush(fout);
  if (fout != stdout) {
    fclose(fout);
  }
  fclose(fin);

  return ret;
}



/*
 * FUNCTION NAME:	convertKanjiCodeOfEachLineInFile
 * DESCRIPTION:		convert kanji code of each line in a file
 * ARGUMENTS:		input kanji code, input file pointer,
 *			output kanji code, output file pointer, mime decode flag
 * RETURN VALUE:	Successful: CONV_OK, otherwise: CONV_ERR
 */
static conv_t
convertKanjiCodeOfEachLineInFile(kin, fin, kout, fout, f_mime)
	kcode_t		kin;
	register FILE *	fin;
	kcode_t		kout;
	register FILE *	fout;
	bool_t		f_mime;
{
  char	buf[BUFSIZ];
  int	err = 0;

  memset(buf, 0, sizeof(buf));
  while (fgets(buf, sizeof(buf), fin) != NULL) {
    char	rbuf[BUFSIZ*2];
    conv_t	ret;

    memset(rbuf, 0, sizeof(rbuf));
    ret = wkfConvertKanjiCodeOfString(kin, buf, kout, rbuf, sizeof(rbuf));
    switch (ret) {
      case CONV_ERR:
	err++;
      case CONV_NO:
	if (f_mime &&
	    wkfDecodeMimeString(buf, rbuf, sizeof(rbuf), kout) == CONV_OK) {
	  fputs(rbuf, fout);
	}
	else {
	  fputs(buf, fout);
	}
	break;
      case CONV_OK:
      default:
	fputs(rbuf, fout);
	break;
    }
    memset(buf, 0, sizeof(buf));
  }

  return (err > 0 ? CONV_ERR : CONV_OK);
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	savePermission
 * DESCRIPTION:		save permission of file
 * ARGUMENTS:		input filename, output filename
 * RETURN VALUE:	none
 */
static void
savePermission(org_name, new_name)
	String	org_name;
	String	new_name;
{
  struct stat		sbuf;

  if (stat(org_name, &sbuf) == -1) {
    fprintf(stderr, "Warning: "
	"%s couldn't set permission on file %s\n", Program, new_name);
    return;
  }

  if (chmod(new_name, sbuf.st_mode) == -1) {
    fprintf(stderr, "Warniing: "
	"%s couldn't set permission on file %s\n", Program, new_name);
    return;
  }
}

/*
 * FUNCTION NAME:	saveTimeStamp
 * DESCRIPTION:		save time-stamp of file
 * ARGUMENTS:		input filename, output filename
 * RETURN VALUE:	none
 */
static void
saveTimeStamp(org_name, new_name)
	String	org_name;
	String	new_name;
{
  struct stat		sbuf;
  struct utimbuf	tbuf;

  if (stat(org_name, &sbuf) == -1) {
    fprintf(stderr, "Warning: "
	"%s couldn't set time-stamp on file %s\n", Program, new_name);
    return;
  }

  tbuf.actime  = sbuf.st_atime;
  tbuf.modtime = sbuf.st_mtime;

  if (utime(new_name, &tbuf) == -1) {
    fprintf(stderr, "Warning: "
	"%s couldn't set time-stamp on file %s\n", Program, new_name);
    return;
  }
}

/*
 * FUNCTION NAME:	tmpName
 * DESCRIPTION:		make a temporary filename
 * ARGUMENTS:		filename
 * RETURN VALUE:	temporary filename
 */
static String
tmpName(fname)
	String	fname;
{
  static char	buf[FILE_LEN];

  memset(buf, 0, sizeof(buf));
  if (snprintf(buf, sizeof(buf) - 1, "%s.wkf~", fname) == -1) {
    return NULL;
  }

  return buf;
}

/*
 * FUNCTION NAME:	backupName
 * DESCRIPTION:		make a backup filename
 * ARGUMENTS:		filename
 * RETURN VALUE:	backup filename
 */
static String
backupName(fname)
	String	fname;
{
  static char	buf[FILE_LEN];

  memset(buf, 0, sizeof(buf));
  if (snprintf(buf, sizeof(buf) - 1, "%s.bak", fname) == -1) {
    return NULL;
  }

  return buf;
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	printErrorMessage
 * DESCRIPTION:		print error message
 * ARGUMENTS:		error type
 * RETURN VALUE:	none
 */
static void
printErrorMessage(err)
	error_t	err;
{
  switch (err) {
    case ERR_NO_CONVERT_OPTION:
      fprintf(stderr,
		"%s: no convert option -- needs 'j' or 'e' or 's' option.\n",
		Program);
      break;
    case ERR_MULTIPLE_CONVERT_OPTION:
      fprintf(stderr,
		"%s: bad option -- j, e, s, c, C selected only one.\n",
		Program);
      break;
    case ERR_MULTIPLE_INPUT_CODE_OPTION:
      fprintf(stderr,
		"%s: bad option -- J, E, S selected only one.\n",
		Program);
      break;
    case ERR_MULTIPLE_LINE_END_OPTION:
      fprintf(stderr,
		"%s: bad option -- M, D, U selected only one.\n",
		Program);
      break;
  }
}

/*
 * FUNCTION NAME:	copyright
 * DESCRIPTION:		print copyright
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
copyright(void)
{
  fprintf(stderr,
	  "%s\n"
	  "Version %s\n"
	  "Copyright (C) 1997-2002 Kouichi ABE (WALL)\n"
	  "%s\n",
	  SOFTWARE_NAME,
	  SOFTWARE_VERSION,
	  SOFTWARE_ADDRESS);
}

/*
 * FUNCTION NAME:	usage
 * DESCRIPTION:		print usage
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
usage(void)
{
  fprintf(stderr,
	  "usage: "
	  "%s [-{j|e|s|c|C}] [-mzALT] [-{J|E|S}] [-{M|D|U}] [<file>...]\n\n",
	  Program);
  exit(1);
}

/*****************************************************************************/

/*
 *	MAIN FUNCTION 
 */
int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  kcode_t	kin;		/* input kanji code */
  kcode_t	kout;		/* output kanji code */
  Option	opt;		/* command option */
  register char	c;

  /* set default value */
  kin		= KC_UNKNOWN;
  kout		= KC_UNKNOWN;
  opt.checkcode	= FALSE;	/* no checking kanji code */
  opt.showfname	= FALSE;	/* no showing filename */
  opt.mimedecode= FALSE;	/* no mime decode */
  opt.backup	= FALSE;	/* no backup */
  opt.savetstamp= FALSE;	/* no saving time-stamp */
  opt.lineconv	= FALSE;	/* no each line converting */
  opt.lineend	= FALSE;	/* default use (UNIX type) */

  /* addressing arguments */
  while ((--argc > 0) && ((*++argv)[0] == '-')) {
    while ((c = *++argv[0]) != EOL) {
      switch (c) {
	case 'h':	/* show usage */
	  copyright();
	  usage();
	case 'v':	/* show version */
	  copyright();
	  exit(0);
	case 'C':	/* check kanji code and show filename */
	  opt.showfname = TRUE;
	case 'c':	/* check kanji code */
	  opt.checkcode = TRUE;
	  if (kin != KC_UNKNOWN || kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  break;
	case 'j':	/* convert to JIS */
	  if (opt.checkcode) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  else if (kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  kout = KC_JIS;
	  break;
	case 'e':	/* convert to Japanese EUC */
	  if (opt.checkcode) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  else if (kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  kout = KC_EUC;
	  break;
	case 's':	/* convert to Shift-JIS */
	  if (opt.checkcode) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  else if (kout != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_CONVERT_OPTION);
	    usage();
	  }
	  kout = KC_SJIS;
	  break;
	case 'm':	/* decode mime string */
	  opt.mimedecode = TRUE;
	  opt.lineconv	 = TRUE;	/* depended */
	  break;
	case 'z':	/* convert 2byte alphabet and figures to 1byte ones */
	  wkfConvertZenkaku2ASCII(TRUE);
	  break;
	case 'A':	/* make a backup file */
	  opt.backup = TRUE;
	  break;
	case 'L':	/* convert string with each line */
	  opt.lineconv = TRUE;
	  break;
	case 'T':	/* save the time-stamp */
	  opt.savetstamp = TRUE;
	  break;
	case 'M':	/* output the line end with Macintosh format */
	  if (opt.lineend) {
	    printErrorMessage(ERR_MULTIPLE_LINE_END_OPTION);
	    usage();
	  }
	  opt.lineend = TRUE;
	  wkfSetLineEndCode(LE_CR);
	  break;
	case 'D':	/* output the line end with DOS format */
	  if (opt.lineend) {
	    printErrorMessage(ERR_MULTIPLE_LINE_END_OPTION);
	    usage();
	  }
	  opt.lineend = TRUE;
	  wkfSetLineEndCode(LE_CRLF);
	  break;
	case 'U':	/* output the line end with UNIX format */
	  if (opt.lineend) {
	    printErrorMessage(ERR_MULTIPLE_LINE_END_OPTION);
	    usage();
	  }
	  opt.lineend = TRUE;
	  wkfSetLineEndCode(LE_LF);
	  break;
	case 'J':	/* input kanji code is JIS */
	  if (kin != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_INPUT_CODE_OPTION);
	    usage();
	  }
	  kin = KC_JIS;
	  break;
	case 'E':	/* input kanji code is Japanese EUC */
	  if (kin != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_INPUT_CODE_OPTION);
	    usage();
	  }
	  kin = KC_EUC;
	  break;
	case 'S':	/* input kanji code is Shift-JIS */
	  if (kin != KC_UNKNOWN) {
	    printErrorMessage(ERR_MULTIPLE_INPUT_CODE_OPTION);
	    usage();
	  }
	  kin = KC_SJIS;
	  break;
	default:	/* illegal option */
	  fprintf(stderr, "%s: illegal option -- %c\n", Program, c);
	  argc = -1;
	  break;
      }
    }
  }


  /* check option to convert kanji code */
  if (kout == KC_UNKNOWN &&	/* no option to convert kanji code */
      opt.checkcode == FALSE) {	/* no check kanji code */
    printErrorMessage(ERR_NO_CONVERT_OPTION);
    usage();
  }

  /* processing every file */
  if (argc > 0) {
    register int	i;

    for (i = 0; i < argc; i++) {
      register FILE *	fin;

      signal(SIGINT,  SIG_IGN);
      fin = fopen(argv[i], "r");
      if (fin == NULL) {
	fprintf(stderr, "failed to open %s\n", argv[i]);
	continue;
      }
      /* only check kanji code */
      if (opt.checkcode) {
	if (opt.showfname) {
	  fprintf(stdout, "%-16s\t", argv[i]);
	  fflush(stdout);
	}
	setvbuf(fin, (char *)NULL, _IOFBF, 0);
	wkfPrintGuessedKanjiCodeOfFile(fin, FALSE);
	continue;
      }
      /* make backup file */
      if (opt.backup) {
	String	tname;
	
	tname = tmpName(argv[i]);
	if (tname != NULL) {
	  register FILE *	fout;
	  
	  fout = fopen(tname, "w");
	  if (fout != NULL) {
	    conv_t	ret;	/* return value of convert function */

	    ret = convertKanjiCodeInFile(kin, fin, kout, fout, &opt);
	    if (ret == CONV_OK) {
	      String	bname;
	      
	      bname = backupName(argv[i]);
	      if (bname != NULL &&
		  rename(argv[i], bname) == 0 &&
		  rename(tname, argv[i]) == 0) {
		if (opt.savetstamp) {		/* save file time-stamp */
		  saveTimeStamp(bname, argv[i]);
		}
		savePermission(bname, argv[i]);	/* save file permission */
	      }
	    }
	    else {
	      unlink(tname);	/* remove the temporary file */
	    }
	  }
	}
	else {
	  fprintf(stderr, "failed to open backup file for %s\n", argv[i]);
	}
      }
      else {
	conv_t	ret;	/* return value of convert function */

	ret = convertKanjiCodeInFile(kin, fin, kout, stdout, &opt);
	if (ret == CONV_ERR) {
	  fprintf(stderr, "%s: no converted ...\n", argv[i]);
	}
      }
      signal(SIGINT,  SIG_DFL);
    }
  }
  else if (argc == 0) {	/* for standard input */
    if (opt.checkcode) {	/* only check kanji code */
      if (opt.showfname) {
	fprintf(stdout, "stdin\t");
	fflush(stdout);
      }
      wkfPrintGuessedKanjiCodeOfFile(stdin, FALSE);
      fflush(stdout);
    }
    else {
      convertKanjiCodeInFile(kin, stdin, kout, stdout, &opt);
    }
  }
  else {	/* show usage */
    usage();
  }

  return 0;
}
